/*----------------------------------------------------------------------*
 * dbinterface.c
 *
 * Michael Geikowski, 16.6.93
 *
 * Implementierung der Schnittstelle zur Datenbasis. Die Datenbasis wird
 * hier auf das UNIX File System abgebildet. Alle Dateien werden als
 * Text-Files gespeichert. Ueber die Umgebungsvariable TLADB kann ein
 * Verzeichnis angegeben werden, in dem die Dateien abgelegt werden.
 * Ist TBRDB nicht definiert, wird das aktuelle Verzeichnis verwendet.
 * Fuer die einzelnen Dateitypen sind folgende Suffixe fuer die
 * Dateinamen vorgesehen:
 *  - graphfile    : .graph
 *  - linkfile     : .link
 *  - contentsfile : .cont
 *  - errorfile    : .err
 *  - bookmarkfile : .mark
 *----------------------------------------------------------------------*/
/* Olaf Meier, 16.6.94
 *
 * Ergaenzungen:
 * -------------
 * Zusaetzliche Dateitypen und ihre Suffixe fuer die Dateinamen:
 * - treefile      : .tree
 * - warningfile   : .warn
 * - vartreefile   : .vars
 *----------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "dbinterface.h"

int standalone = 0;	// 1, wenn Standalone Modus, sonst 0.
int lineno;		// Muss beim Aufruf von SimpleError() und
			// SimpleErrorCont() die aktuelle Zeilennummer des
                        // gescannten Quelltextes enthalten.

static FILE* graphfile    = 0;	// Die FilePointer
static FILE* linkfile     = 0;
static FILE* errfile      = 0;
static FILE* contfile     = 0;
static FILE* bookmarkfile = 0;
static FILE* treefile     = 0;
static FILE* warningfile  = 0;
static FILE* vartreefile  = 0;

static int  errors;		// Anzahl ausgegebener Fehler
static int  errlistlength;	// Anzahl Zeilen der Fehlerliste
				// ( eine Fehlermeldung kann mehrere Zeilen
				//   benoetigen )
static int warnings;            // Anzahl ausgegebener Warnungen
static int warnlistlength;      // Anzahl Zeilen der Warnungsliste
                                // ( eine Warnung kann mehrere Zeilen
                                //   benoetigen )

static char lastchar;		// letztes in Fehlerdatei geschriebenes Zeichen
static char lastwarnchar;       // letztes in Warnungsdatei geschriebenes Zeichen
static int  contlistlength;	// Anzahl Zeilen eines Inhaltsverzeichnisses
static long maxlinkpos;		// max. Startposition eines Links im aktuell
				// gelesenen LinkFile
static int  graphwidth,		// Groesse des aktuell gelesenen Graphs
            graphheight;

static int  vartreewidth,       // Groesse des aktuell gelesenen Systemvariablen-Baums
            vartreeheight; 

       
static char errmode,contmode,	// Lese/Schreibmodus von errfile/contfile/
            warnmode,           // warnfile
            bookmarkmode;       // bookmarkfile
static int  bookmarkcount;	// Anzahl Lesezeichen einer Datei

/* -----Hilfsfunktionen-------------------------------------------------*/

static char* ReadID(FILE* infile)
	// Bezeichner bel. Laenge aus Datei lesen
 {
 const int bufinc = 255;
 char      *ID,*buf,*tempbuf;
 int       i = 0, buflen = bufinc;

 buf = new char[buflen];
					// fuehrende Leerzeichen ueberlesen
 while( (buf[0] = fgetc(infile)) == ' ') ;

 while( (buf[i] != ' ')  &&
        (buf[i] != '\n') &&
        (! feof(infile))   )
   {
   i++;
   buf[i] = fgetc(infile);
   if (i == buflen-2)			// Puffer bei Ueberlauf vergroessern
     {
     buf[buflen-1] = '\0';
     buflen += bufinc;
     tempbuf = new char[buflen];
     strcpy(tempbuf,buf);
     delete buf;
     buf = tempbuf;
     }
   }
 buf[i] = '\0';
 ID = new char[strlen(buf)+1];
 strcpy(ID,buf);
 delete buf;
 return ID;
 }

static char* ReadLine(FILE* infile)
	// Zeile bel. Laenge aus Datei lesen
 {
 const int bufinc = 255;
 char      *line,*buf,*tempbuf;
 int       i = 0, buflen = bufinc;

 buf = new char[buflen];

 buf[0] = fgetc(infile);

 while( (buf[i] != '\n') && (! feof(infile)) )
   {
   i++;
   buf[i] = fgetc(infile);
   if (i == buflen-2)			// Puffer bei Ueberlauf vergroessern
     {
     buf[buflen-1] = '\0';
     buflen += bufinc;
     tempbuf = new char[buflen];
     strcpy(tempbuf,buf);
     delete buf;
     buf = tempbuf;
     }
   }
 buf[i] = '\0';
 line = new char[strlen(buf)+1];
 strcpy(line,buf);
 delete buf;
 return line;
 }

/*----------------------------------------------------------------------*
 * OpenGraphFile
 *
 * Oeffne eine Graph-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * width und height geben die Groesse eines zu schreibenden  Graphen an.
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/

int OpenGraphFile(char* filename, char* mode, int width, int height)
 {
 char* outfilename;
 char* dir;

 if (standalone) return 1;
 assert(graphfile==0 && filename && mode && mode[1]=='\0' &&
        ( mode[0]=='w' || mode[0]=='r')                     );
 dir = getenv("TBRDB");
 if (dir)
   {
   outfilename = new char[strlen(dir)+strlen(filename)+8];
   sprintf(outfilename,"%s/%s.graph",dir,filename);
   }
 else
   {
   outfilename = new char[strlen(filename)+7];
   sprintf(outfilename,"%s.graph",filename);
   }
 graphfile = fopen(outfilename,mode);
 if ( graphfile )
   {
   if ( mode[0] == 'w' )
     {
     fprintf(graphfile,"%d %d\n",width,height);
     }
   else
     {
     if (fscanf(graphfile,"%d %d\n",&graphwidth,&graphheight) != 2)
       {
       fclose(graphfile);
       graphfile=0;
       }
     }
   }
 delete outfilename;
 return (graphfile != 0);
 }

/*--------------------------------------------------------------------*
 * OpenVarTreeFile
 * Oeffne eine Systemvariablen-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w").
 * Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * width und height geben die Groesse eines zu schreibenden Graphen an.
 * Liefert True bei Erfolg, False sonst.
 *--------------------------------------------------------------------*/

int OpenVarTreeFile(char* filename, char* mode, int width, int height)
  {
 char* outfilename;
 char* dir;
 if (standalone) return 1;
 assert(vartreefile==0 && filename && mode && mode[1]=='\0' &&
        ( mode[0]=='w' || mode[0]=='r')                     );
 dir = getenv("TBRDB");
 if (dir)
   {
   outfilename = new char[strlen(dir)+strlen(filename)+7];
   sprintf(outfilename,"%s/%s.vars",dir,filename);
   }
 else
   {
   outfilename = new char[strlen(filename)+6];
   sprintf(outfilename,"%s.vars",filename);
   }
 vartreefile = fopen(outfilename,mode);
 if ( vartreefile )
   {
   if ( mode[0] == 'w' )
     {
     fprintf(vartreefile,"%d %d\n",width,height);
     }
   else
     {
     if (fscanf(vartreefile,"%d %d\n",&vartreewidth,&vartreeheight) != 2)
       {
       fclose(vartreefile);
       vartreefile=0;
       }
     }
   }
 delete outfilename;
 return (vartreefile != 0);
 }

/*----------------------------------------------------------------------*
 *  OpenParseTreeFile
 *
 * Oeffne eine Parsetree-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * width und height geben die Groesse eines zu schreibenden Parsetree-Graphen an.
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/

int OpenParseTreeFile(char* filename, char* mode)
{
 char* outfilename;
 char* dir;

 if (standalone) return 1;
 assert(treefile==0 && filename && mode && mode[1]=='\0' &&
        ( mode[0]=='w' || mode[0]=='r')                     );
 dir = getenv("TBRDB");
 if (dir)
   {
   outfilename = new char[strlen(dir)+strlen(filename)+7];
   sprintf(outfilename,"%s/%s.tree",dir,filename);

   }
 else
   {
   outfilename = new char[strlen(filename)+6];
   sprintf(outfilename,"%s.tree",filename);
   }
 treefile = fopen(outfilename,mode);
 
 delete outfilename;
 return (treefile != 0);
 }




/*----------------------------------------------------------------------*
 * GetGraphExtension
 *
 * Liefert Breite und Hoehe des aktuell gelesenen Graphen.
 *----------------------------------------------------------------------*/

void GetGraphExtension(int *width, int *height)
 {
 assert(standalone==0);
 *width  = graphwidth;
 *height = graphheight;
 }


/*----------------------------------------------------------------------*
 * GetVarTreeExtension
 *
 * Liefert Breite und Hoehe des aktuell gelesenen Variablenbaums.
 *----------------------------------------------------------------------*/

void GetVarTreeExtensions(int *width, int *height)
 {
 assert(standalone==0);
 *width  = vartreewidth;
 *height = vartreeheight;
 }

/*----------------------------------------------------------------------*
 * ReadNode
 *
 * Lese einen Knoten mit Typ `t', Name `name' und Koordinaten (x,y)
 * aus aktueller Graph-Datei.
 * Alle von dem Knoten ausgehenden Kanten muessen nach dem ReadNode()
 * Aufruf durch ReadEdge() gelesen werden.
 * 'endofnodes' ist True, wenn kein Knoten mehr in der Datei enthalten
 * ist.
 * Ergibt False, bei Lesefehlern.
 *----------------------------------------------------------------------*/

 int ReadNode(char* t, char** name, int* x, int* y, int* endofnodes)
  {
  char c;
  int  ok;
  
  assert(standalone==0);
  assert(graphfile!=0);
  
  c = fgetc(graphfile);
  if (feof(graphfile))
    {
    *endofnodes = 1;
    return 1;
    }
  *endofnodes = 0;
  ok    = (c == '#');
  *t    = fgetc(graphfile);
  *name = ReadID(graphfile);
  ok    = ok && ( fscanf(graphfile,"%d %d\n",x,y) == 2 );
  return ok;
  }


/*----------------------------------------------------------------------*
 * ReadVarNode
 *
 * Lese einen Knoten mit Typ `t', Name `name' und Koordinaten (x,y)
 * aus aktueller Systemvariablen-Datei.
 * Alle von dem Knoten ausgehenden Kanten muessen nach dem ReadNode()
 * Aufruf durch ReadEdge() gelesen werden.
 * 'endofnodes' ist True, wenn kein Knoten mehr in der Datei enthalten
 * ist.
 * Ergibt False, bei Lesefehlern.
 *----------------------------------------------------------------------*/

 int ReadVarNode(char* t, char** name, int* x, int* y, int* endofnodes)
 {
  char c;
  int  ok;

  assert(standalone==0);
  assert(vartreefile!=0);

  c = fgetc(vartreefile);
  if (feof(vartreefile))
    {
    *endofnodes = 1;
    return 1;
    }
  *endofnodes = 0;
  ok    = (c == '#');
  *t    = fgetc(vartreefile);
  *name = ReadID(vartreefile);
  ok    = ok && ( fscanf(vartreefile,"%d %d\n",x,y) == 2 );
  return ok;
  }

/*----------------------------------------------------------------------*
 * ReadPTNode
 *
 * Lese einen Knoten mit Typ `t', Name `name'
 * und Subtreepositionen (mixpos, maxpos) aus aktueller ParseTree-Graph-Datei.
 * Alle von dem Knoten ausgehenden Kanten muessen nach dem ReadPTNode()
 * Aufruf durch ReadPTEdge() gelesen werden.
 * 'endofnodes' ist True, wenn kein Knoten mehr in der Datei enthalten
 * ist.
 * Ergibt False, bei Lesefehlern.
 *----------------------------------------------------------------------*/

int ReadPTNode(char* t,char** name,  long* minpos, long* maxpos, int* endofnodes)
 {
  char c;
  int  ok;

  assert(standalone==0);
  assert(treefile!=0);

  c = fgetc(treefile);
  if (feof(treefile))
    {
    *endofnodes = 1;
    return 1;
    }
  *endofnodes = 0;
  ok    = (c == '#');
  *t    = fgetc(treefile);
  *name = ReadID(treefile);
  ok    = ok && ( fscanf(treefile,"%ld %ld\n",minpos,maxpos) == 2 );
  return ok;
  }

/*----------------------------------------------------------------------*
 * ReadEdge
 *
 * Lese eine Kante aus der aktuellen Graph-Datei, die vom letzten 
 * gelesenen Knoten ausgeht. Die Kante hat Typ `t' und fuehrt zum 
 * Knoten `name'.
 * 'endofedges' ist True, wenn keine Kante mehr von dem Knoten ausgeht.
 * Ergibt False, bei Lesefehlern.
 *----------------------------------------------------------------------*/

 int ReadEdge(char* t, char** name, int* endofedges)
  {
  char c;
  
  assert(standalone==0);
  assert(graphfile!=0);

  c = fgetc(graphfile);
  if (c != '>')
    {
    *endofedges = 1;
    ungetc(c,graphfile);
    return 1;
    }
  *endofedges = 0;
  *t    = fgetc(graphfile);
  *name = ReadID(graphfile);
  return 1;
  }



/*----------------------------------------------------------------------*
 * ReadVarEdge
 *
 * Lese eine Kante aus der aktuellen Systemvariablen-Datei, die vom letzten
 * gelesenen Knoten ausgeht. Die Kante hat Typ `t' und fuehrt zum
 * Knoten `name'.
 * 'endofedges' ist True, wenn keine Kante mehr von dem Knoten ausgeht.
 * Ergibt False, bei Lesefehlern.
 *----------------------------------------------------------------------*/

 int ReadVarEdge(char* t, char** name, int* endofedges)
 {
  char c;

  assert(standalone==0);
  assert(vartreefile!=0);

  c = fgetc(vartreefile);
  if (c != '>')
    {
    *endofedges = 1;
    ungetc(c,vartreefile);
    return 1;
    }
  *endofedges = 0;
  *t    = fgetc(vartreefile);
  *name = ReadID(vartreefile);
  return 1;
  }


/*----------------------------------------------------------------------*
 *
 * Lese die naechste Aktionsnummer des aktuellen Variablenknotens
 * aus der Systemvariablen-Datei.
 * 'endofactions' ist True, wenn der aktuelle Knoten keine weitere
 * Aktionsnummer enthaelt.
 *---------------------------------------------------------------------*/

int ReadVarAction(int *action, int* endofactions)
 {
  char c;
  int ok;
  assert(standalone==0);
  assert(vartreefile!=0);
  c = fgetc(vartreefile);
  if (c != '*')
    {
    *endofactions = 1;
    ungetc(c,vartreefile);
    return 1;
    }
  *endofactions = 0;
  ok = (fscanf(vartreefile,"%d\n",action) == 1);
  return 1;
 }

/*----------------------------------------------------------------------*
 *  ReadPTEdge
 *
 * Lese eine Kante aus der aktuellen ParseTree-Graph-Datei, die vom letzten
 * gelesenen Knoten ausgeht. Die Kante hat Typ `t' , fuehrt zum
 * Knoten `name' und besitzt die Subtreepositionen(min,max).
 * 'endofedges' ist True, wenn keine Kante mehr von dem Knoten ausgeht.
 * Ergibt False, bei Lesefehlern.
 *----------------------------------------------------------------------*/

int ReadPTEdge(char* t, char** name, long* min, long* max, int* endofedges)
  {
   char c;
   int ok;
  assert(standalone==0);
  assert(treefile!=0);

  c = fgetc(treefile);
  if (c != '>')
    {
    *endofedges = 1;
    ungetc(c,treefile);
    return 1;
    }
  *endofedges = 0;
  *t    = fgetc(treefile);
  *name = ReadID(treefile);
  ok = (fscanf(treefile,"%ld %ld\n",min,max) == 2); 
  return 1;
  }

/*----------------------------------------------------------------------*
 * WriteNode
 *
 * Schreibe einen Knoten mit Typ `t', Name `name' und Koordinaten (x,y)
 * in aktuelle Graph-Datei.
 * Alle von dem Knoten ausgehenden Kanten muessen nach dem WriteNode()
 * Aufruf durch WriteEdge() geschrieben werden.
 *----------------------------------------------------------------------*/

 void WriteNode(char t, char* name, int x, int y)
  {
  if (standalone) return;
  assert(graphfile!=0 && name);
  fprintf(graphfile,"#%c %s %d %d\n",t,name,x,y);
  }


/*----------------------------------------------------------------------*
 * WriteVarNode
 *
 * Schreibe einen Variablen-Knoten mit Typ `t', Name `name' und Koordinaten (x,y)
 * in aktuelle Systemvariablen-Datei.
 * Alle von dem Knoten ausgehenden Kanten muessen nach dem WriteVarNode()
 * Aufruf durch WriteVarEdge() geschrieben werden.
 * Die Aktionen des aktuellen Knotens muessen mit WriteVarAction() geschrieben
 * werden.
 *----------------------------------------------------------------------*/

 void WriteVarNode(char t, char* name, int x, int y)
  {
  if (standalone) return;
  assert(vartreefile!=0 && name);
  fprintf(vartreefile,"#%c %s %d %d\n",t,name,x,y);
  }

/*----------------------------------------------------------------------*
 * WritePTNode
 *
 * Schreibe einen Knoten mit Typ `t', Name `name' 
 * und Subtreepositionen (minpos, maxpos)  in aktuelle ParseTree-Graph-Datei.
 * Alle von dem Knoten ausgehenden Kanten muessen nach dem WritePTNode()
 * Aufruf durch WritePTEdge() geschrieben werden.
 *----------------------------------------------------------------------*/

 void WritePTNode(char t, char* name, long minpos, long maxpos)
  {
  if (standalone) return;
  assert(treefile!=0 && name);
  fprintf(treefile,"#%c %s %ld %ld\n",t,name,minpos,maxpos);
  }

/*----------------------------------------------------------------------*
 * WriteEdge
 *
 * Schreibe eine Kante in die aktuelle Graph-Datei, die vom letzten in
 * die Datei geschriebenen Knoten ausgeht. Die Kante hat Typ `t' und
 * fuehrt zum Knoten `name'.
 *----------------------------------------------------------------------*/

 void WriteEdge(char t, char* name)
  {
  if (standalone) return;
  assert(graphfile!=0 && name);
  fprintf(graphfile,">%c %s\n",t,name);
  }



/*----------------------------------------------------------------------*
 * WriteVarEdge
 *
 * Schreibe eine Kante in die aktuelle Systemvariablen-Datei, die vom letzten in
 * die Datei geschriebenen Knoten ausgeht. Die Kante hat Typ `t' und
 * fuehrt zum Knoten `name'.
 *----------------------------------------------------------------------*/

 void WriteVarEdge(char t, char* name)
  {
  if (standalone) return;
  assert(vartreefile!=0 && name);
  fprintf(vartreefile,">%c %s\n",t,name);
  }


/*-----------------------------------------------------------------------*
 * WriteVarAction
 *
 * Schreibe eine Aktionsnummer in die aktuelle Systemvariablen-Datei, die
 * zum letzten in die Datei geschriebenen Variablenknoten gehoert.
 *-----------------------------------------------------------------------*/

 void WriteVarAction(int action)
  {
  if (standalone) return;
  assert(vartreefile!=0 && (action >= 0));
  fprintf(vartreefile,"*%d\n",action);
  }


       

/*----------------------------------------------------------------------*
 * WritePTEdge
 *
 * Schreibe eine Kante in die aktuelle ParseTree-Graph-Datei, die vom letzten in
 * die Datei geschriebenen Knoten ausgeht. Die Kante hat Typ `t' und
 * fuehrt zum Knoten `name' und hat die Subtreepositionen (min, max).
 *----------------------------------------------------------------------*/

 void WritePTEdge(char t, char* name, long min, long max)
  {
   if (standalone) return;
  assert(treefile!=0 && name);
  fprintf(treefile,">%c %s %ld %ld\n",t,name,min,max);
  }

/*----------------------------------------------------------------------*
 * CloseGraphFile
 *
 * Schliesse die aktuelle Graph-Datei.
 *----------------------------------------------------------------------*/

void CloseGraphFile()
 {
 if (standalone) return;
 assert(graphfile!=0);
 fclose(graphfile);
 graphfile=0;
 }


/*----------------------------------------------------------------------*
 * CloseVarTreeFile
 *
 * Schliesse die aktuelle Systemvariablen-Datei.
 *----------------------------------------------------------------------*/

void CloseVarTreeFile()
{
 if (standalone) return;
 assert(vartreefile!=0);
 fclose(vartreefile);
 vartreefile=0;
 }

/*----------------------------------------------------------------------*
 * CloseParseTreeFile
 *
 * Schliesse die aktuelle ParseTree-Graph-Datei.
 *----------------------------------------------------------------------*/

void CloseParseTreeFile()
 {
  if (standalone) return;
  assert(treefile!=0);
  fclose(treefile);
  treefile=0;
 }

/*----------------------------------------------------------------------*
 * OpenLinkFile
 *
 * Oeffne eine Link-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * maxpos gibt die groesste Position eines Link Startpunktes an, falls
 * `mode' = "w".
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/

int OpenLinkFile(char* filename, char* mode, long maxpos)
 {
 char* outfilename;
 char* dir;
 char  c;
 
 if (standalone) return 1;
 assert(linkfile==0 && filename && mode && mode[1]=='\0' &&
        ( mode[0]=='w' || mode[0]=='r')                     );
 dir = getenv("TBRDB");
 if (dir)
   {
   outfilename = new char[strlen(dir)+strlen(filename)+7];
   sprintf(outfilename,"%s/%s.link",dir,filename);
   }
 else
   {
   outfilename = new char[strlen(filename)+6];
   sprintf(outfilename,"%s.link",filename);
   }
 linkfile = fopen(outfilename,mode);
 if ( linkfile )
   {
   if ( mode[0] == 'w' )
     {
     fprintf(linkfile,"%ld\n",maxpos);
     }
   else
     {
     if (fscanf(linkfile,"%ld\n",&maxlinkpos) != 1)
       {
       fclose(linkfile);
       linkfile=0;
       }
     else
       {
       c = fgetc(linkfile);
       ungetc(c,linkfile);
       }
     }
   }
 delete outfilename;
 return (linkfile != 0);
 }

/*----------------------------------------------------------------------*
 * GetMaxLinkPos()
 *
 * Wurde die Link-Datei zum Lesen geoeffnet, gibt diese Funktion die
 * maximale Startposition unter allen Verweisen der Datei aus.
 *----------------------------------------------------------------------*/

int GetMaxLinkPos()
 {
 assert(standalone==0);
 return maxlinkpos;
 }
 
/*----------------------------------------------------------------------*
 * ReadLink
 *
 * Lese einen Link aus der aktuellen Link-Datei. `pos' und `leng' geben
 * Position und Laenge des Link Startsymbols, `text' gibt Link-Text an.
 * Es folgen die Zielpunkte des Links. Speicherplatz fuer `...modname'
 * und `text' wird bereitgestellt, bzw. `*...modname' ist NULL, wenn
 * entsprechende Position gleich -1.
 * Ergibt True, wenn fehlerfrei.
 *----------------------------------------------------------------------*/

int ReadLink(long *pos,int *leng,
             char **text,
             long *defpos , char **defmodname,
             long *prevpos, char **prevmodname,
             long *nextpos, char **nextmodname,
             long *modpos , char **destmodname)
 {
 int ok;
 
 assert(standalone==0);
 assert(linkfile != 0);
 *defmodname = *prevmodname = *nextmodname = *destmodname = 0;
 ok = ( fscanf(linkfile,"# %ld %d\n#",pos,leng) == 2 );
 *text=ReadLine(linkfile);
 ok = ok && (fscanf(linkfile,"> %ld",defpos) == 1);
 if (ok && (*defpos != -1))
   *defmodname = ReadID(linkfile);
 ok = ok && (fscanf(linkfile,"%ld",prevpos) == 1);
 if (ok && (*prevpos != -1))
   *prevmodname = ReadID(linkfile);
 ok = ok && (fscanf(linkfile,"%ld",nextpos) == 1);
 if (ok && (*nextpos != -1))
   *nextmodname = ReadID(linkfile);
 ok = ok && (fscanf(linkfile,"%ld",modpos) == 1);
 if (ok && (*modpos != -1))
   *destmodname = ReadID(linkfile);
 else
   ok = ok && (fgetc(linkfile) == '\n');
 char c = fgetc(linkfile);
 ungetc(c,linkfile);
 return ok;
 }
 
/*----------------------------------------------------------------------*
 * EndOfLinkFile()
 *
 * Ergibt True, wenn Link-Datei zum Lesen geoeffnet wurde und alle
 * Links durch ReadLink gelesen wurden.
 *----------------------------------------------------------------------*/

int EndOfLinkFile()
 {
 assert(standalone==0);
 return feof(linkfile);
 }
 
/*----------------------------------------------------------------------*
 * WriteLink
 *
 * Schreibe einen neuen Link in die aktuelle Link-Datei. `pos' und `leng'
 * geben Position und Laenge des Symbols, das Ausgangspunkt des Links ist.
 * `s1', `s2' und `s3' werden konkateniert und ergeben den Link-Text.
 * Es folgen die Zielpunkte des Links. Wird ein Zielpunkt nicht
 * genutzt, muss sein Name gleich NULL sein.
 *----------------------------------------------------------------------*/

void WriteLink(long pos,int leng,
               char* s1, char* s2, char* s3,
               long defpos , char* defmodname,
               long prevpos, char* prevmodname,
               long nextpos, char* nextmodname,
               long modpos , char* destmodname)
 {
 if (standalone) return;
 assert(linkfile!=0 && s1 && s2 && s3);
 fprintf(linkfile,"# %ld %d\n",pos,leng);
 fprintf(linkfile,"#%s%s%s\n",s1,s2,s3);
 if (defmodname)
   fprintf(linkfile,"> %ld %s ",defpos,defmodname);
 else
   fprintf(linkfile,"> -1 ");
 if (prevmodname)
   fprintf(linkfile,"%ld %s ",prevpos,prevmodname);
 else
   fprintf(linkfile,"-1 ");
 if (nextmodname)
   fprintf(linkfile,"%ld %s ",nextpos,nextmodname);
 else
   fprintf(linkfile,"-1 ");
 if (destmodname)
   fprintf(linkfile,"%ld %s\n",modpos,destmodname);
 else
   fprintf(linkfile,"-1\n");
 }

/*----------------------------------------------------------------------*
 * CloseLinkFile
 *
 * Schliesse die aktuelle Link-Datei.
 *----------------------------------------------------------------------*/

void CloseLinkFile()
 {
 if (standalone) return;
 assert(linkfile!=0);
 fclose(linkfile);
 linkfile=0;
 }

/*----------------------------------------------------------------------*
 * OpenErrorFile
 *
 * Oeffne eine Error-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/

int OpenErrorFile(char* filename, char* mode)
 {
 char* outfilename;
 char* dir;
 
 errlistlength = errors = 0;
 lastchar = '\n';
 if (standalone)
   {
   errmode  = 'w';
   return 1;
   }
 assert(errfile==0 && filename && mode && mode[1]=='\0' &&
        ( mode[0]=='w' || mode[0]=='r')                     );
 errmode = mode[0];
 dir = getenv("TBRDB");
 if (dir)
   {
   outfilename = new char[strlen(dir)+strlen(filename)+6];
   sprintf(outfilename,"%s/%s.err",dir,filename);
   }
 else
   {
   outfilename = new char[strlen(filename)+5];
   sprintf(outfilename,"%s.err",filename);
   }
 errfile = fopen(outfilename,mode);
 if ( errfile )
   {
   if (mode[0]=='r')
     {
     if ( (fseek(errfile,-10,/* SEEK_END */ 2) == -1) ||
          (fscanf(errfile,"%d",&errlistlength) !=  1) ||
          (fseek(errfile,0,/* SEEK_SET */ 0)   == -1)   )
       {
       fclose(errfile);
       errfile=0;
       }
     }
   }
 delete outfilename;
 return (errfile != 0);
 }


/*----------------------------------------------------------------------*
 * OpenWarningFile
 *
 * Oeffne eine Warning-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/
 
int OpenWarningFile(char* filename, char* mode)
  {
 char* outfilename;
 char* dir;

 warnlistlength = warnings = 0;
 lastwarnchar = '\n';
 if (standalone)
   {
   warnmode  = 'w';
   return 1;
   }
 assert(warningfile==0 && filename && mode && mode[1]=='\0' &&
        ( mode[0]=='w' || mode[0]=='r')                     );
 warnmode = mode[0];
 dir = getenv("TBRDB");
 if (dir)
   {
   outfilename = new char[strlen(dir)+strlen(filename)+7];
   sprintf(outfilename,"%s/%s.warn",dir,filename);
   }
 else
  {
   outfilename = new char[strlen(filename)+6];
   sprintf(outfilename,"%s.warn",filename);
   }
 warningfile = fopen(outfilename,mode);
 if ( warningfile )
   {
   if (mode[0]=='r')
     {
     if ( (fseek(warningfile,-10,/* SEEK_END */ 2) == -1) ||
          (fscanf(warningfile,"%d",&warnlistlength) !=  1) ||
          (fseek(warningfile,0,/* SEEK_SET */ 0)   == -1)   )
       {
       fclose(warningfile);
       warningfile=0;
       }
     }
   }
 delete outfilename;
 return (warningfile != 0);
 }


/*----------------------------------------------------------------------*
 * GetNrOfErrors
 *
 * Wurde die Error-Datei zum Lesen geoeffnet, gibt diese Funktion die
 * Anzahl der Zeilen, die in der Datei enthalten sind, aus.
 * (Eine Fehlermeldung kann sich ueber mehrere Zeilen erstrecken)
 *----------------------------------------------------------------------*/

int GetNrOfErrors()
 {
 assert(standalone==0);
 return errlistlength;
 }


/*----------------------------------------------------------------------*
 * GetNrOfWarnings
 *
 * Wurde die Warningr-Datei zum Lesen geoeffnet, gibt diese Funktion die
 * Anzahl der Zeilen, die in der Datei enthalten sind, aus.
 * (Eine Warnungsmeldung kann sich ueber mehrere Zeilen erstrecken)
 *----------------------------------------------------------------------*/
             

int GetNrOfWarnings()
 { 
 assert(standalone==0);
 return warnlistlength;
 } 

/*----------------------------------------------------------------------*
 * SimpleError
 *
 * Schreibe eine Fehlermeldung `msg' (ohne Newline) in aktuelle Fehler-
 * Datei mit Fehlerort `errpos', `errleng'. Soll die Meldung mehrzeilig
 * werden, kann sie mit SimpleErrorCont() fortgesetzt werden.
 * Ueber die globale Variable `standalone' kann das Schreiben in die Datei
 * verhindert werden. Statt dessen wird die Meldung auf stderr ausgegeben.
 *----------------------------------------------------------------------*/

void SimpleError(char *msg, long errpos, int errleng)
 {
 errors++;
 errlistlength++;
 lastchar='\n';
 if (standalone)
   {
   if (lastchar != '\n')
     { 
     fprintf(stderr,"\n");
     }
   fprintf(stderr,"at line %d: %s\n", lineno, msg);
   fflush(stderr);
   return;
   }
 assert(errfile!=0 && msg);
 if (lastchar != '\n')
   { 
   fprintf(errfile,"\n> a -1 0\n");
   errlistlength++;
   }
 fprintf(errfile,"#at line %d: %s\n", lineno, msg);
 fprintf(errfile,"> a %ld %d\n", errpos, errleng);
 }

/*----------------------------------------------------------------------*
 * SimpleErrorCont
 *
 * Setzt die letzte Fehlermeldung fort ( siehe SimpleError() ) um die
 * Zeichenkette `msg'. Das Ende einer Zeile wird durch Newline
 * gekennzeichnet. Die letzte Zeile wird beim naechsten Aufruf von
 * SimpleError() oder CloseErrorFile() automatisch abgeschlossen.
 * Ueber die globale Variable `standalone' kann das Schreiben in die Datei
 * verhindert werden. Statt dessen wird die Meldung auf stderr ausgegeben.
 *----------------------------------------------------------------------*/

void SimpleErrorCont(char *msg)
 {
 char *p;

 if (standalone)
   {
   lastchar = msg[strlen(msg)-1];
   fprintf(stderr,msg);
   fflush(stderr);
   return;
   }
 assert(errfile && msg);
 p = msg;
 while (*p != '\0')
   {
   if (lastchar=='\n') fputc('#',errfile);
   fputc(*p,errfile);
   if (*p=='\n')
     { 
     fprintf(errfile,"> a -1 0\n");
     errlistlength++;
     }
   lastchar=*p;
   p++;
   }
 }


/*----------------------------------------------------------------------*
 * SimpleWarning
 *
 * Schreibe eine Warnung `msg' (ohne Newline) in aktuelle Warnungs-
 * Datei mit Ort `errpos', `errleng'. Soll die Meldung mehrzeilig
 * werden, kann sie mit SimpleWarningCont() fortgesetzt werden.
 * Ueber die globale Variable `standalone' kann das Schreiben in die Datei
 * verhindert werden. Statt dessen wird die Meldung auf stderr ausgegeben.
 *----------------------------------------------------------------------*/

void SimpleWarning(char *msg, long warnpos, int warnleng, int line)
 {
 warnings++;
 warnlistlength++;
 lastwarnchar='\n';
 if (standalone)
   {
   if (lastwarnchar != '\n')
     {
     fprintf(stderr,"\n");
     }
   fprintf(stderr,"at line %d: %s\n", line, msg);
   fflush(stderr);
   return;
   }
 assert(warningfile!=0 && msg);
 if (lastwarnchar != '\n')
   {
   fprintf(warningfile,"\n> a -1 0\n");
   warnlistlength++;
   }
 fprintf(warningfile,"#at line %d: %s\n", line, msg);
 fprintf(warningfile,"> a %ld %d\n", warnpos, warnleng);
}


/*----------------------------------------------------------------------*
 * SimpleWarningCont
 *
 * Setzt die letzte Warnungsmeldung fort ( siehe SimpleWarning() ) um die
 * Zeichenkette `msg'. Das Ende einer Zeile wird durch Newline
 * gekennzeichnet. Die letzte Zeile wird beim naechsten Aufruf von
 * SimpleWarning() oder CloseWarningFile() automatisch abgeschlossen.
 * Ueber die globale Variable `standalone' kann das Schreiben in die Datei
 * verhindert werden. Statt dessen wird die Meldung auf stderr ausgegeben.
 *----------------------------------------------------------------------*/

void SimpleWarningCont(char *msg)
 {
 char *p;

 if (standalone)
   {
   lastwarnchar = msg[strlen(msg)-1];
   fprintf(stderr,msg);
   fflush(stderr);
   return;
   }
 assert(warningfile && msg);
 p = msg;
 while (*p != '\0')
   {
   if (lastwarnchar=='\n') fputc('#',warningfile);
   fputc(*p,warningfile);
   if (*p=='\n')
     {
     fprintf(warningfile,"> a -1 0\n");
     warnlistlength++;
     }
   lastwarnchar=*p;
   p++;
   }
 }

/*----------------------------------------------------------------------*
 * ReadErrorLine
 *
 * Lese eine Zeile `line' aus der Fehler-Datei mit Fehlerort
 * `pos', `leng'.
 * Ergibt False bei Lesefehler.
 *----------------------------------------------------------------------*/

int ReadErrorLine(char** line, long* pos, int* leng)
 {
 char c;
 int ok;
 
 assert(standalone==0);
 assert(errfile!=0);
 c = fgetc(errfile); 
 if ( feof(errfile) || (c!='#') )
   {
   return 0;
   }
 *line = ReadLine(errfile);
 ok = ( fscanf(errfile,"> a %ld %d\n",pos,leng) == 2 );
 return ok;
 }

/*----------------------------------------------------------------------*
 * ReadWarningLine
 *
 * Lese eine Zeile `line' aus der Warnungs-Datei mit Ort der Warnung
 * `pos', `leng'.
 * Ergibt False bei Lesefehler.
 *----------------------------------------------------------------------*/

int ReadWarningLine(char** line, long *pos, int* leng)
 {
 char c;
 int ok;

 assert(standalone==0);
 assert(warningfile!=0);
 c = fgetc(warningfile);
 if ( feof(warningfile) || (c!='#') )
   {
   return 0;
   }
 *line = ReadLine(warningfile);
 ok = ( fscanf(warningfile,"> a %ld %d\n",pos,leng) == 2 );
 return ok;
 }

/*----------------------------------------------------------------------*
 * CloseErrorFile
 *
 * Schliesse die aktuelle Error-Datei.
 * Gibt die Anzahl geschriebener Fehlermeldungen zurueck (nur bei
 * mode = "w").
 *----------------------------------------------------------------------*/

int CloseErrorFile()
 {
 if (standalone)
   {
   if (lastchar != '\n')
     { 
     fprintf(stderr,"\n");
     fflush(stderr);
     }
   return errors;
   }
 assert(errfile!=0);
 if ( (errmode == 'w') && (lastchar != '\n') )
   { 
   fprintf(errfile,"\n> a -1 0\n");
   errlistlength++;
   }
 if (errmode == 'w')
   {
   if (errors==0)
     {
     fprintf(errfile,"#no errors\n> a 0 0\n");
     errlistlength++;
     }
   fprintf(errfile,".          %d\n",errlistlength);
   }
 fclose(errfile);
 errfile=0;
 return(errors);
 }


/*----------------------------------------------------------------------*
 * CloseWarningFile
 *
 * Schliesse die aktuelle Warnungs-Datei.
 * Gibt die Anzahl geschriebener Warnungsmeldungen zurueck (nur bei
 * mode = "w").
 *----------------------------------------------------------------------*/

int CloseWarningFile()
 {
 if (standalone)
   {
   if (lastwarnchar != '\n')
     {
     fprintf(stderr,"\n");
     fflush(stderr);
     }
   return warnings;
   }
 assert(warningfile!=0);
 if ( (warnmode == 'w') && (lastwarnchar != '\n') )
   {
   fprintf(warningfile,"\n> a -1 0\n");
   warnlistlength++;
   }
 if (warnmode == 'w')
   {
   if (warnings==0)
     {
     fprintf(warningfile,"#no warnings\n> a 0 0\n");
     warnlistlength++;
     }
   fprintf(warningfile,".          %d\n",warnlistlength);
   }
 fclose(warningfile);
 warningfile=0;
 return(warnings);
 }

/*----------------------------------------------------------------------*
 * OpenContentsFile
 *
 * Oeffne eine Inhaltsverzeichnis-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/

int OpenContentsFile(char* filename, char* mode)
 {
 char* outfilename;
 char* dir;

 if (standalone) return 1; 
 assert(contfile==0 && filename && mode && mode[1]=='\0' &&
        ( mode[0]=='w' || mode[0]=='r')                     );
 contmode = mode[0];
 dir = getenv("TBRDB");
 if (dir)
   {
   outfilename = new char[strlen(dir)+strlen(filename)+7];
   sprintf(outfilename,"%s/%s.cont",dir,filename);
   }
 else
   {
   outfilename = new char[strlen(filename)+6];
   sprintf(outfilename,"%s.cont",filename);
   }
 contfile = fopen(outfilename,mode);
 if ( contfile )
   {
   contlistlength = 0;
   if (mode[0]=='r')
     {
     if ( (fseek(contfile,-10, /* SEEK_END */ 2) == -1) ||
          (fscanf(contfile,"%d",&contlistlength) !=  1) ||
          (fseek(contfile,0,/* SEEK_SET */ 0)    == -1)   )
       {
       fclose(contfile);
       contfile=0;
       }
     }
   }
 delete outfilename;
 return (contfile != 0);
 }

/*----------------------------------------------------------------------*
 * GetNrOfContentsLines
 *
 * Wurde die Inhalts-Datei zum Lesen geoeffnet, gibt diese Funktion die
 * Anzahl der Zeilen des gespeicherten Inhaltsverzeichnisses aus.
 *----------------------------------------------------------------------*/

int GetNrOfContentsLines()
 {
 assert(standalone==0);
 return contlistlength;
 }
 
/*----------------------------------------------------------------------*
 * BeginContentsLine
 *
 * Schreibe eine Zeile in die Inhaltsverzeichnis-Datei. Die Zeile wird
 * `depth' weit eingerueckt, stueckweise mit AddToContentsLine()
 * geschrieben und mit EndContentsLine() abgeschlossen.
 *----------------------------------------------------------------------*/

void BeginContentsLine(int depth)
 {
 if (standalone) return;
 assert(contfile!=0);
 contlistlength++;
 fprintf(contfile,"#");
 for (int i=1; i<depth; i++)
   fprintf(contfile,"  ");
 }

/*----------------------------------------------------------------------*
 * AddToContentsLine
 *
 * Fuege eine weitere Zeichenkette (ohne Space und Newline) zur aktuellen
 * Zeile hinzu. Die Zeichenketten werden automatisch durch Space getrennt.
 * ( siehe BeginContentsLine() )
 *----------------------------------------------------------------------*/

void AddToContentsLine(char *s)
 {
 if (standalone) return;
 assert(contfile!=0 && s);
 fprintf(contfile,"%s ",s);
 }

/*----------------------------------------------------------------------*
 * EndContentsLine
 *
 * Schliesse aktuelle Zeile ab mit Sprungziel `pos', `leng'.
 * ( siehe BeginContentsLine() )
 *----------------------------------------------------------------------*/

void EndContentsLine(long pos, int leng)
 {
 if (standalone) return;
 assert(contfile!=0);
 fprintf(contfile,"\n> a %ld %d\n",pos,leng);
 }

/*----------------------------------------------------------------------*
 * WriteContentsLine
 *
 * Schreibe eine Zeile in die Inhaltsverzeichnis-Datei mit Sprungziel
 * `pos', `leng'. Die Zeile ergibt sich aus Konkatenation von `s1' und
 * `s2' (`s2' kann auch NULL sein) und wird `depth' weit eingerueckt.
 *----------------------------------------------------------------------*/

void WriteContentsLine(int depth, char* s1, char* s2, long pos, int leng)
 {
 if (standalone) return ;
 assert(contfile!=0 && s1);
 contlistlength++;
 fprintf(contfile,"#");
 for (int i=1; i<depth; i++)
   fprintf(contfile,"  ");
 if (s2)
   fprintf(contfile,"%s %s ",s1,s2);
 else
   fprintf(contfile,"%s ",s1);
 fprintf(contfile,"\n> a %ld %d\n",pos,leng);
 }

/*----------------------------------------------------------------------*
 * ReadContentsLine
 *
 * Lese eine Zeile `line' aus der Inhaltsverzeichnis-Datei mit Sprungziel
 * `pos', `leng'.
 * Ergibt False bei Lesefehler.
 *----------------------------------------------------------------------*/

int ReadContentsLine(char** line, long* pos, int* leng)
 {
 char c;
 int ok;
 
 assert(standalone==0); 
 assert(contfile!=0);
 c = fgetc(contfile); 
 if ( feof(contfile) || (c!='#') )
   {
   return 0;
   }
 *line = ReadLine(contfile);
 ok = ( fscanf(contfile,"> a %ld %d\n",pos,leng) == 2 );
 return ok;
 }
 
/*----------------------------------------------------------------------*
 * CloseContentsFile
 *
 * Schliesse die aktuelle Inhaltsverzeichnis-Datei.
 *----------------------------------------------------------------------*/

void CloseContentsFile()
 {
 if (standalone) return;
 assert(contfile!=0);
 if (contmode == 'w')
   {
   fprintf(contfile,".          %d\n",contlistlength);
   }
 fclose(contfile);
 contfile=0;
 }

/*----------------------------------------------------------------------*
 * OpenBookmarkFile
 *
 * Oeffne eine Lesezeichen-Datei mit Namen `filename' zum
 * lesen (`mode'="r") oder schreiben (`mode'="w"). Im letzten Fall wird
 * eine evtl. unter diesem Namen bereits existierende Datei geloescht.
 * Liefert True bei Erfolg, False sonst.
 *----------------------------------------------------------------------*/

int OpenBookmarkFile(char* filename, char* mode)
 {
 char* outfilename;
 char* dir;

 if (standalone) return 1; 
 assert(bookmarkfile==0 && filename && mode && mode[1]=='\0' &&
        ( mode[0]=='w' || mode[0]=='r')                     );
 bookmarkmode = mode[0];
 dir = getenv("TBRDB");
 if (dir)
   {
   outfilename = new char[strlen(dir)+strlen(filename)+7];
   sprintf(outfilename,"%s/%s.mark",dir,filename);
   }
 else
   {
   outfilename = new char[strlen(filename)+6];
   sprintf(outfilename,"%s.mark",filename);
   }
 bookmarkfile = fopen(outfilename,mode);
 if ( bookmarkfile )
   {
   bookmarkcount = 0;
   if (mode[0]=='r')
     {
     if ( (fseek(bookmarkfile,-10, /* SEEK_END */ 2) == -1) ||
          (fscanf(bookmarkfile,"%d",&bookmarkcount)  !=  1) ||
          (fseek(bookmarkfile,0,/* SEEK_SET */ 0)    == -1)   )
       {
       fclose(bookmarkfile);
       bookmarkfile=0;
       }
     }
   }
 delete outfilename;
 return (bookmarkfile != 0);
 }

/*----------------------------------------------------------------------*
 * GetNrOfBookmarks
 *
 * Wurde die Lesezeichen-Datei zum Lesen geoeffnet, gibt diese Funktion die
 * Anzahl der gespeicherten Lesezeichen aus.
 *----------------------------------------------------------------------*/

int GetNrOfBookmarks()
 {
 assert(standalone==0);
 return bookmarkcount;
 }
 
/*----------------------------------------------------------------------*
 * WriteBookmark
 *
 * Schreibe ein Lesezeichen in die Lesezeichen-Datei mit Sprungziel
 * `ModName', `pos' und Name 'name'.
 *----------------------------------------------------------------------*/

void WriteBookmark(char* name, char* ModName, long pos)
 {
 if (standalone) return ;
 assert(bookmarkfile!=0 && name && ModName);
 bookmarkcount++;
 fprintf(bookmarkfile,"#%s\n> %s %ld 0\n",name,ModName,pos);
 }

/*----------------------------------------------------------------------*
 * ReadBookmark
 *
 * Lese ein Lesezeichen aus der Lesezeichen-Datei mit Sprungziel
 * `ModName', `pos' und Name 'name'.
 * Ergibt False bei Lesefehler.
 *----------------------------------------------------------------------*/

int ReadBookmark(char** name, char** ModName, long* pos)
 {
 char c;
 int ok;
 
 assert(standalone==0); 
 assert(bookmarkfile!=0);
 c = fgetc(bookmarkfile); 
 if ( feof(bookmarkfile) || (c!='#') )
   {
   return 0;
   }
 *name = ReadLine(bookmarkfile);
 c = fgetc(bookmarkfile); 
 if ( feof(bookmarkfile) || (c!='>') )
   {
   delete name;
   return 0;
   }
 *ModName = ReadID(bookmarkfile);
 ok = ( fscanf(bookmarkfile,"%ld 0\n",pos) == 1 );
 return ok;
 }

/*----------------------------------------------------------------------*
 * CloseBookmarkFile
 *
 * Schliesse die aktuelle Lesezeichen-Datei.
 *----------------------------------------------------------------------*/

void CloseBookmarkFile()
 {
 if (standalone) return;
 assert(bookmarkfile!=0);
 if (bookmarkmode == 'w')
   {
   fprintf(bookmarkfile,".          %d\n",bookmarkcount);
   }
 fclose(bookmarkfile);
 bookmarkfile=0;
 }
