/*----------------------------------------------------------------------*
 * ParseTree.h
 *
 * Olaf Meier, 16.6.94
 *
 *  Definiert die Klassen ParseExpr und ParseSymb sowie die Hilfsklassen
 *  SmStack und SmStackElem.
 *  Sie dienen der Erzeugung eines Ableitungsbaumes pro Spezifikationsmodul.
 *  Die Klasse ParseExpr dient der Darstellung einer Produktionsregel innerhalb
 *  der Grammatik fuer die tTLA+ Syntax. Jedes dieser Objekte besitzt wiederum
 *  Objekte der Klasse ParseSymb, die jeweils ein Terminalsymbol bzw. Nichtterminal-
 *  symbol (Symbole auf der rechten Seite der Produktionsregel.
 * 
 *  Die Objekte der Klassen SmStack und SmStackElem bilden einen Stack
 *  zur temporaeren Speicherung von Nichtterminalsymbolen, da der Ableitungsbaum
 *  Bottom-up erzeugt wird.
 *  
 *----------------------------------------------------------------------------*/

#ifndef PARSETREE_H
#define PARSETREE_H


#include <stdio.h>
#include "list.h"
#include "symlist.h"
#include "adjlistgnu.h" // dient zur Konvertierung der Parsetree-Baumstruktur
                    // in eine Adjaszenslisten-Darstellung.

enum Symbol_Type { NonTerm, Terminal, Empty};  // Flags der Klasse ParseSymb:
                                              // NonTerm = NichtTerminalSymbol
                                              // Terminal = Terminalsymbol
                                              // Empty= Epsilon-Produktionsregel

// Flags der Klasse ParseExpr: Interner Repraesentation der  verschiedenen Produktionsregeln

enum noTerminal_Type {tunknown, tOneModule, tModuleType, tModuleName01, tHeaderParts,
                      tHeaderPart, tExport_ID , tExport_qualifierPrefixes,
                      tExport_ID_csl, tRestexport, tBodyParts, tBodyPart, tCM_Definition,
                      tTM_Definition, tModuleEnd, tImportList, tIncludeEntry, tIncludeParts,
                      tIncludePart, tIncludeParameters01, tSubstitutionPart01, tSubstitution_,
                      tOneSubstitution_, tConstantDefinitions, tBooleanDefinitions,
                      tPredicateDefinitions, tStateFunctionDefinitions, tActionDefinitions,
                      tTransitionFunctionDefinitions, tAssumptionDefinitions, tTheoremDefinitions,
                      tTemporalDefinitions, tConstantDefinition, tBooleanDefinition,
                      tPredicateDefinition, tStateFunctionDefinition, tActionDefinition,
                      tTransitionFunctionDefinition, tAssumptionDefinition, tTheoremDefinition,
                      tTemporalDefinition, tConstantExpr, tBooleanExpr, tStateFcnExpr, tPredExpr,
                      tTransFctExpr, tActionExpr, tconstExpr_csl, texpr_csl, texpr, tSetExpr,
                      tbeginSet, texpr_list, tpredefSet, tsetRel, tsetOp, tBinOp, tStringExpr,
                      tisOrNotElementOf, trecdef_csl, trecdef, trctype_ssl, trctype, tboolRel,
                      tboolRelIB, tBinRel, tOperatorApplication, tprefixLines, tconjunctLines,
                      tdisjunctLines, tisElementOfSetListQB, tCaseRA_list, tCaseRA, tSymbolDef_LHS,
                      tformalPar_ssl, tformalPar, tinfix, tTypeAndSortDecl01, tTypeAndSortDecl,
                      tTypeDecl, tSort, tModuleName, tqualifyingID, tcomponentID, tID,
                      tqualifierPrefixes, tnormalID_csl, tID_csl, 
                      expanded, substituted, corrected };

class ParseExpr;

// Die Klasse ParseSymb beinhaltet die Referenz auf ein Objekt der Klasse ParseExpr oder
// ein Objekt der Klasse SymListElem (Normal Identifier aus der Symboltabelle.
// Die Kennung `st' dient dem typentsprechenden Zugriff.

class ParseSymb 

{
   friend class ParseExpr;

   public:

   ParseSymb()
    : st(Empty)
   {
     termsymb = NULL;
     notermsymb = NULL;
    };
   

   
   ~ParseSymb();
   

   
   Symbol_Type GetSymbolType(); // Rueckgabe des Symboltyps: `Terminal' fuer SymListElem 
                                //                         `NonTerm' fuer ParseExpr
                                //                         `Empty'   fuer keine Referenz auf Objekt

    SymListElem* GetSymbElem(); // Rueckgabe des SymListElem-Elements


    ParseExpr* GetParseExpr();  // Rueckgabe des ParseExpr-Elements



   private:

   Symbol_Type st;  // Flag: Symboltype


   SymListElem *termsymb; // Entweder SymListElem oder

   ParseExpr *notermsymb; // ParseExpr

};

// Die Klasse ParseExpr dient der Repraesentation der einzelnen Produktionsregeln.
// Sie besitzt eine Referenz auf ein variables Array von ParseSymb-Objekten.
// Fuer spaetere Traversierungen des Ableitungsbaums ist der maximale Index des Arrays
// vermerkt. Die Symbolliste `slist' wird in der Typueberpruefungsphase benutzt, um
// eine evtl. Aenderung der Typinformationen pro includiertem Bezeichner vorzunehmen.

class ParseExpr 

  {
     public:
     long minpos; // Minimale und maximale Position der Symbole des durch das ParseExpr-Objekt
     long maxpos; // aufgespannten Teilbaums, im jeweiligen Spezifikationsmoduls.
      


     ParseExpr(noTerminal_Type tp,char* name,int i);

     ~ParseExpr();

     void AddExpr(ParseExpr *pe,int i);  // Hinzufuegen eines ParseExpr-Objekts mit Arrayindex `i'

     void AddExpr(SymListElem *se,int i); // Hinzufuegen eines SymListElem-Objekts mit Arrayindex `i'

     int GetMaxIndex(); // Rueckgabe des maximalen Arrayindex diese ParseExpr-Objekts

     void AddIncludes(SymList* inclist); // Hinzufuegen einer Bezeichnerliste, die in das Modul dieses
                                         // Parsetrees includiert werden.  

     char* GetRuleName();                // Rueckgabe des Knotennamens (Produktionsregel-Namen)

     SymList* GetIncludes();             // Rueckgabe der Includeliste

     ParseSymb* GetExpr(int i);   // Rueckgabe des Array-Objekts mit Index `i'

     ParseExpr* GetPE(int i);     // Rueckgabe des ParseExpr-Objekt, das im Array-Objekt mit
                                  // Index `i' abgelegt ist.

     SymListElem* GetSE(int i);   // Rueckgabe des SymListElem-Objekts, das im Array-Objekt mit
                                  // Index `i' abgelegt ist.

     noTerminal_Type GetNonTermType(); // Rueckgabe des Nichterminal-Symbolflags des aktuellen Objekts

    
    
    private:

    int maxindex; // Maximal ansprechbarer Arrayindex in dem ParseExpr-Objekt
    
    noTerminal_Type nt_type;  // Flag : Non Terminal-Typ

    char* ruleName;          // Name der Produktionsregel

    ParseSymb *parsesymbols; // Referenz auf ein ParseSymbPointer-Array 

    SymList *slist;   // Includeliste bei Include-Expression 
 };


// Semantischer Stack zum temporaeren Zwischenspeichern von Nichtterminalsymbolen:

class SmStackElem : public ListElem

     {
        friend class SmStack;
        public:

        SmStackElem(ParseExpr* pe )
         : pexpptr(pe)
         {}
        
        ~SmStackElem()
         {}
   
        private:

        ParseExpr* pexpptr;  // Referenz auf die ge`pop'te Produktionsregel
     };

class SmStack : public List
     {
        public:

        SmStack(SmStackElem *e =  0)
          : List(e)
         {}
        
        ~SmStack()
         {}

        int SmStackIsEmpty(); // Rueckgabe, ob Stack leer (==1) 

        void SmStackPush(SmStackElem *se); // Push-Operation 

        ParseExpr* SmStackPop(); // Pop-Operation 

        void SmStackClear(); // Loeschung aller vorhandenen Elemente des Stacks
     };

// Non-Member Funktionen


// `PopandAppend' fuegt das ge`pop'te ParseExpr-Objekt in das Array des 'pexplist'-Objekt  
// unter Index `i' ein.
              
void PopandAppend(ParseExpr *pexp, ParseExpr *pexplist, int i);

// `AppendtoList' fuegt das Terminalsymbol `sle' in das Array des 'pexplist'-Objekt
// unter Index `i' ein.

void AppendtoList(SymListElem *sle, ParseExpr *pexplist ,char* toNa,int i);




// `TraverseTree' fuehrt ein DFS-Baumdurchlauf im Baum `tree' aus.
// Dient nur zu Testzwecken.

void TraverseTree(ParseExpr* tree);


// Hilfsfunktion zur Umwandlung der ParseTree-Baumstruktur in eine Adjaszenslisten-Darstellung
// Ableitungsbaum 'tree'  mit Wurzelknoten 'currnode' wird in 'adj' dargestellt.

void ConverttoAdjList(ParseExpr* tree,Node* currnode, AdjList* adj);



// Umwandlung des Parsetrees in eine Adjaszenslisten-Darstellung und 
//  Abspeichern dieser Liste in Datei 'filename.tree'

void WriteParsetree(ParseExpr* tree,Node* currnode, AdjList* adj,char* filename);


#endif
